"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Scalar(google.protobuf.message.Message):
    """A Scalar represents data that does not fulfill the promise of a Dataset.
    A Dataset promises to have a schema and the possibility to iterate on
    pyarrow.RecordBatches
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class Spec(google.protobuf.message.Message):
        """Definitions
        How to obtain the dataset
        """
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        TRANSFORMED_FIELD_NUMBER: builtins.int
        MODEL_FIELD_NUMBER: builtins.int
        PRIVACY_PARAMS_FIELD_NUMBER: builtins.int
        RANDOM_SEED_FIELD_NUMBER: builtins.int
        SYNTHETIC_MODEL_FIELD_NUMBER: builtins.int
        @property
        def transformed(self) -> global___Scalar.Transformed: ...
        @property
        def model(self) -> global___Scalar.Model: ...
        @property
        def privacy_params(self) -> global___Scalar.PrivacyParameters: ...
        @property
        def random_seed(self) -> global___Scalar.RandomSeed: ...
        @property
        def synthetic_model(self) -> global___Scalar.SyntheticModel: ...
        def __init__(self,
            *,
            transformed : typing.Optional[global___Scalar.Transformed] = ...,
            model : typing.Optional[global___Scalar.Model] = ...,
            privacy_params : typing.Optional[global___Scalar.PrivacyParameters] = ...,
            random_seed : typing.Optional[global___Scalar.RandomSeed] = ...,
            synthetic_model : typing.Optional[global___Scalar.SyntheticModel] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["model",b"model","privacy_params",b"privacy_params","random_seed",b"random_seed","spec",b"spec","synthetic_model",b"synthetic_model","transformed",b"transformed"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["model",b"model","privacy_params",b"privacy_params","random_seed",b"random_seed","spec",b"spec","synthetic_model",b"synthetic_model","transformed",b"transformed"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal["spec",b"spec"]) -> typing.Optional[typing_extensions.Literal["transformed","model","privacy_params","random_seed","synthetic_model"]]: ...

    class Transformed(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class NamedArgumentsEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: typing.Text = ...
            value: typing.Text = ...
            def __init__(self,
                *,
                key : typing.Text = ...,
                value : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

        TRANSFORM_FIELD_NUMBER: builtins.int
        ARGUMENTS_FIELD_NUMBER: builtins.int
        NAMED_ARGUMENTS_FIELD_NUMBER: builtins.int
        transform: typing.Text = ...
        """Transform id"""

        @property
        def arguments(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
            """Dataset or other object ids"""
            pass
        @property
        def named_arguments(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
        def __init__(self,
            *,
            transform : typing.Text = ...,
            arguments : typing.Optional[typing.Iterable[typing.Text]] = ...,
            named_arguments : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["arguments",b"arguments","named_arguments",b"named_arguments","transform",b"transform"]) -> None: ...

    class Model(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class PretrainedModel(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            FOUNDATION_MODEL_NAME_FIELD_NUMBER: builtins.int
            PEFT_INITIALISATION_URI_FIELD_NUMBER: builtins.int
            foundation_model_name: typing.Text = ...
            peft_initialisation_uri: typing.Text = ...
            def __init__(self,
                *,
                foundation_model_name : typing.Text = ...,
                peft_initialisation_uri : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["foundation_model_name",b"foundation_model_name","peft_initialisation_uri",b"peft_initialisation_uri"]) -> None: ...

        PRETRAINED_MODEL_FIELD_NUMBER: builtins.int
        @property
        def pretrained_model(self) -> global___Scalar.Model.PretrainedModel: ...
        def __init__(self,
            *,
            pretrained_model : typing.Optional[global___Scalar.Model.PretrainedModel] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["pretrained_model",b"pretrained_model","type",b"type"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["pretrained_model",b"pretrained_model","type",b"type"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal["type",b"type"]) -> typing.Optional[typing_extensions.Literal["pretrained_model"]]: ...

    class PrivacyParameters(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class Point(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            EPSILON_FIELD_NUMBER: builtins.int
            DELTA_FIELD_NUMBER: builtins.int
            epsilon: builtins.float = ...
            delta: builtins.float = ...
            def __init__(self,
                *,
                epsilon : builtins.float = ...,
                delta : builtins.float = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["delta",b"delta","epsilon",b"epsilon"]) -> None: ...

        POINTS_FIELD_NUMBER: builtins.int
        @property
        def points(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Scalar.PrivacyParameters.Point]: ...
        def __init__(self,
            *,
            points : typing.Optional[typing.Iterable[global___Scalar.PrivacyParameters.Point]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["points",b"points"]) -> None: ...

    class RandomSeed(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        VALUE_FIELD_NUMBER: builtins.int
        value: builtins.int = ...
        def __init__(self,
            *,
            value : builtins.int = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["value",b"value"]) -> None: ...

    class SyntheticModel(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    UUID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    DOC_FIELD_NUMBER: builtins.int
    SPEC_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    uuid: typing.Text = ...
    """A Scalar does not ensure this possibility. As a consequence, oprations
    from standard libraries are allowed (pandas.mean, numpy.std,...) but
    operations implemented for Datasets by Sarus like computing marginals or
    fitting a Keras model cannot be performed on a Scalar.

    Scalars are generated by transforms that explicitly require a specific
    format (e.g. as_pandas, as_numpy,...) or as byproducts of transforms
    (model weights, training history,...).

    e.g. RFC 4122 id used to refer to the dataset (content linked?)
    """

    name: typing.Text = ...
    doc: typing.Text = ...
    @property
    def spec(self) -> global___Scalar.Spec: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """Other properties"""
        pass
    def __init__(self,
        *,
        uuid : typing.Text = ...,
        name : typing.Text = ...,
        doc : typing.Text = ...,
        spec : typing.Optional[global___Scalar.Spec] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["spec",b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["doc",b"doc","name",b"name","properties",b"properties","spec",b"spec","uuid",b"uuid"]) -> None: ...
global___Scalar = Scalar



from __future__ import annotations

import typing as t
import numpy as np

from sarus_data_spec import typing as st
from sarus_data_spec.path import straight_path
from sarus_data_spec.constants import DATA, OPTIONAL_VALUE, PU_COLUMN

try:
    import sarus_synthetic_data.configs.global_config as sd_config
    import sarus_synthetic_data.configs.typing as sd_typing
    from sarus_synthetic_data.correlations_generator.jax_model.codecs.categorical import (
        CategoricalBuilder,
    )
    from sarus_synthetic_data.correlations_generator.jax_model.codecs.dynamic_array import (
        DynamicArrayBuilder,
    )
    from sarus_synthetic_data.correlations_generator.jax_model.codecs.optional import (
        OptionalBuilder,
    )
    from sarus_synthetic_data.correlations_generator.jax_model.codecs.struct import (
        StructBuilder,
    )
    from sarus_synthetic_data.correlations_generator.jax_model.codecs.text import (
        TextBuilder,
    )
except ModuleNotFoundError:
    pass
import pyarrow as pa
import os


def update_config(
    config_with_noise: sd_config.SyntheticConfig,
    synthetic_data_dir: str,
    parent_stats: st.Statistics,
    links_stats: t.List[st.LinkStatistics],
) -> sd_config.SyntheticConfig:
    config_with_noise.saving_directory = synthetic_data_dir
    for table_name, table_config in config_with_noise.tables.items():
        table_path = straight_path([DATA, *table_name])
        table_stat = parent_stats.nodes_statistics(table_path)[0]
        current_saving_dir = os.path.join(synthetic_data_dir, *table_name)
        update_table_config(table_stat, table_config, current_saving_dir)

    links_list = [
        sd_config.LinkInfo(
            primary_key=tuple(
                link_statistics.pointed().to_strings_list()[0][1:]
            ),
            foreign_key=tuple(
                link_statistics.pointing().to_strings_list()[0][1:]
            ),
            count_distribution=sd_config.IntegerDistribution(
                values=link_statistics.distribution().values(),  # type:ignore[arg-type]
                probabilities=link_statistics.distribution().probabilities(),
            ),
        )
        for link_statistics in links_stats
    ]
    if len(links_list) > 0:
        config_with_noise.links = sd_config.LinksConfig(
            links_info_list=links_list, seed=2
        )
    return config_with_noise


def update_table_config(
    table_stat: st.Statistics,
    table_config: sd_config.TableConfig,
    current_saving_dir: str,
) -> sd_config.TableConfig:
    """Methods updates:
    - each of the correlation/independent config (see respective method for how)
    -
    """
    if table_config.independent_generation is not None:
        table_config.independent_generation.sampling_config.total_size = (
            table_stat.size()
        )
        update_values_and_probabilities_config(
            table_stat, table_config.independent_generation
        )

    if table_config.correlation_generation is not None:
        correlation_config = update_correlation_config(
            table_stat, table_config.correlation_generation, current_saving_dir
        )
        table_config.correlation_generation = correlation_config
    return table_config


def update_values_and_probabilities_config(
    table_stat: st.Statistics,
    config: t.Union[sd_config.IndependentConfig, sd_config.CorrelationConfig],
) -> None:
    """Method that updates the following fields of the config:
    - the number of lines to to be sampled for the table
    - the distribution of each column
    """

    stat_fields = table_stat.children()

    for column_name, col_config in config.columns.items():
        curr_stat = stat_fields[column_name]
        if isinstance(
            col_config, sd_config.OptionalIndependentColumn
        ) or isinstance(col_config, sd_config.OptionalCorrelationColumn):
            size_all = curr_stat.size()
            size_non_null = curr_stat.children()[OPTIONAL_VALUE].size()
            col_config.distribution = type(col_config.distribution)(
                values=[0, 1],
                probabilities=[
                    1 - size_non_null / size_all,
                    size_non_null / size_all,
                ],
            )
            child_col = col_config.child_col
            if not isinstance(child_col, sd_config.SimpleColumn):
                distrib = curr_stat.children()[OPTIONAL_VALUE].distribution()
                if isinstance(child_col, sd_config.ColumnWithStrDistribution):
                    child_col.distribution = type(child_col.distribution)(
                        values=child_col.distribution.values,
                        probabilities=distrib.probabilities(),
                    )
                else:
                    child_col.distribution = type(child_col.distribution)(
                        values=distrib.values(),  # type:ignore[arg-type]
                        probabilities=distrib.probabilities(),
                    )

        elif not isinstance(col_config, sd_config.SimpleColumn):
            distrib = curr_stat.distribution()
            if isinstance(col_config, sd_config.ColumnWithStrDistribution):
                col_config.distribution = type(col_config.distribution)(
                    values=col_config.distribution.values,
                    probabilities=distrib.probabilities(),
                )
            else:
                col_config.distribution = type(col_config.distribution)(
                    values=distrib.values(),  # type:ignore[arg-type]
                    probabilities=distrib.probabilities(),
                )


def update_correlation_config(
    table_stat: st.Statistics,
    correlation_config: sd_config.CorrelationConfig,
    current_saving_dir: str,
) -> sd_config.CorrelationConfig:
    correlation_config.training.check_pointing.output_dir = current_saving_dir
    correlation_config.sampling.total_size = table_stat.size()
    update_values_and_probabilities_config(
        table_stat=table_stat, config=correlation_config
    )
    stat_fields = table_stat.children()
    sub_builders = []

    for name, col_config in correlation_config.columns.items():
        builders = builder_from_col_config(name, col_config, stat_fields[name])
        sub_builders.extend(builders)

    struct = StructBuilder(
        name="table",
        sub_builders=sub_builders,
        n_heads=4,
        n_blocks=1,
        use_pretrained=False,
        load_pretrained_dir="",
        save_trained=False,
        save_trained_dir="",
    )
    # for DP
    max_mult = int(table_stat.multiplicity())
    list_builder = DynamicArrayBuilder(
        name=PU_COLUMN,
        value_builder=struct,
        batch_size=min(10, max_mult),
        max_size=max_mult,
        n_heads=4,
        n_blocks=1,
        use_pretrained=False,
        load_pretrained_dir="",
        save_trained=False,
        save_trained_dir="",
    )
    correlation_config.model.codec_builder.sub_builder = list_builder
    return correlation_config


def builder_from_col_config(
    name: str,
    col_config: t.Union[
        sd_config.OptionalCorrelationColumn,
        sd_config.NonOptionalCorrelationColumn,
    ],
    stats: st.Statistics,
) -> t.List[t.Any]:
    if isinstance(col_config, sd_config.OptionalCorrelationColumn):
        return [
            OptionalBuilder(
                sub_builder=el,
                name=name,
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            )
            for el in builder_from_col_config(
                OPTIONAL_VALUE,
                col_config.child_col,
                stats.children()[OPTIONAL_VALUE],
            )
        ]

    col_type = col_config.col_type
    distribution_kind = col_config.distribution_kind

    if (
        col_type == sd_typing.TypeKind.Datetime
        and distribution_kind == sd_typing.DistributionKind.quantiles
    ):
        vals = col_config.distribution.values
        min_year = pa.compute.year(
            pa.scalar(vals[0], pa.timestamp("ns"))
        ).as_py()
        max_year = pa.compute.year(
            pa.scalar(vals[-1], pa.timestamp("ns"))
        ).as_py()

        return [
            CategoricalBuilder(
                vocab_size=max_year - min_year + 1,
                name="year",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
            CategoricalBuilder(
                vocab_size=12,
                name="month",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
            CategoricalBuilder(
                vocab_size=31,
                name="day",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
            CategoricalBuilder(
                vocab_size=24,
                name="hour",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
            CategoricalBuilder(
                vocab_size=60,
                name="minutes",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
            CategoricalBuilder(
                vocab_size=60,
                name="seconds",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
        ]
    elif (
        col_type == sd_typing.TypeKind.Date
        and distribution_kind == sd_typing.DistributionKind.quantiles
    ):
        vals = col_config.distribution.values
        min_year = pa.compute.year(
            pa.scalar(np.int32(vals[0]), pa.date32())
        ).as_py()
        max_year = pa.compute.year(
            pa.scalar(np.int32(vals[-1]), pa.date32())
        ).as_py()

        return [
            CategoricalBuilder(
                vocab_size=max_year - min_year + 1,
                name="year",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
            CategoricalBuilder(
                vocab_size=12,
                name="month",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
            CategoricalBuilder(
                vocab_size=31,
                name="day",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
        ]

    elif (
        col_type == sd_typing.TypeKind.Time
        and distribution_kind == sd_typing.DistributionKind.quantiles
    ):
        return [
            CategoricalBuilder(
                vocab_size=24,
                name="hour",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
            CategoricalBuilder(
                vocab_size=60,
                name="minutes",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
            CategoricalBuilder(
                vocab_size=60,
                name="seconds",
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            ),
        ]

    elif isinstance(col_config, sd_config.TextCorrelationCol):
        return [
            TextBuilder(
                name=name,
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
                n_tokens=10,
                max_length=t.cast(
                    int, min(stats.distribution().values()[-1], 100)
                ),  # TODO: this is number of char, should be number of tokens
                hidden_size=64,  # TinyStories
            )
        ]
    else:
        vals = col_config.distribution.values
        if (
            col_config.distribution_kind
            == sd_typing.DistributionKind.quantiles
            and vals[0] == vals[1]
        ):
            # for quantiles there are always at least 2 values
            vocab_size = len(vals) - 1
        else:
            vocab_size = len(vals)
        return [
            CategoricalBuilder(
                vocab_size=vocab_size,
                name=name,
                use_pretrained=False,
                load_pretrained_dir="",
                save_trained=False,
                save_trained_dir="",
            )
        ]



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import numpy.typing as npt
import pandas as pd

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusParameterArray,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from scipy.sparse import spmatrix
    from sklearn import model_selection
    from sklearn.base import BaseEstimator
    from sklearn.model_selection import BaseCrossValidator, GridSearchCV
except ModuleNotFoundError:
    BaseEstimator = Any
    BaseCrossValidator = Any
    spmatrix = Any
    GridSearchCV = Any


class sk_kfold(ExternalOpImplementation):
    _transform_id = "sklearn.SK_KFOLD"
    _signature = SarusSignature(
        SarusParameter(
            name="n_splits",
            annotation=int,
            default=5,
        ),
        SarusParameter(
            name="shuffle",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return model_selection.KFold(**kwargs)


class sk_repeated_stratified_kfold(ExternalOpImplementation):
    _transform_id = "sklearn.SK_REPEATED_STRATIFIED_KFOLD"
    _signature = SarusSignature(
        SarusParameter(
            name="n_splits",
            annotation=int,
            default=5,
        ),
        SarusParameter(
            name="n_repeats",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return model_selection.RepeatedStratifiedKFold(**kwargs)


class sk_cross_val_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_CROSS_VAL_SCORE"
    _signature = SarusSignature(
        SarusParameter(
            name="estimator",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=npt.ArrayLike,
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="groups",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="scoring",
            annotation=Optional[Union[str, Callable]],
            default=None,
        ),
        SarusParameter(
            name="cv",
            annotation=Optional[Union[int, BaseCrossValidator, Iterable]],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="fit_params",
            annotation=Optional[Dict[str, Any]],
            default=None,
        ),
        SarusParameter(
            name="pre_dispatch",
            annotation=Optional[Union[str, int]],
            default="2*n_jobs",
        ),
        SarusParameter(
            name="error_score",
            annotation=Union[Literal["raise"], np.number],
            default=np.nan,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return model_selection.cross_val_score(**kwargs)


class sk_train_test_split(ExternalOpImplementation):
    _transform_id = "sklearn.SK_TRAIN_TEST_SPLIT"
    _signature = SarusSignature(
        SarusParameterArray(
            name="arrays",
            annotation=Sequence[
                Union[List, np.ndarray, spmatrix, pd.DataFrame]
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="test_size",
            annotation=Optional[Union[float, int]],
            default=None,
        ),
        SarusParameter(
            name="train_size",
            annotation=Optional[Union[float, int]],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="shuffle",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="stratify",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        arrays, kwargs = signature.collect()
        return model_selection.train_test_split(*arrays, **kwargs)


class sk_time_series_split(ExternalOpImplementation):
    _transform_id = "sklearn.SK_TIME_SERIES_SPLIT"
    _signature = SarusSignature(
        SarusParameter(
            name="n_splits",
            annotation=int,
            default=5,
        ),
        SarusParameter(
            name="max_train_size",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="test_size",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="gap",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return model_selection.TimeSeriesSplit(**kwargs)


class SkGridSearchCV(ExternalOpImplementation):
    _transform_id = "sklearn.SK_GRID_SEARCH_CV"
    _signature = SarusSignature(
        SarusParameter(
            name="estimator",
            annotation=BaseEstimator,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="param_grid",
            annotation=Union[Dict[str, Any], List[Dict[str, Any]]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="scoring",
            annotation=Optional[Union[str, Callable, List, Tuple, Dict]],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="refit",
            annotation=Union[bool, str, Callable],
            default=True,
        ),
        SarusParameter(
            name="cv",
            annotation=Optional[Union[int, BaseCrossValidator, Iterable]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="pre_dispatch",
            annotation=Union[str, int],
            default="2*n_jobs",
        ),
        SarusParameter(
            name="error_score",
            annotation=Union[Literal["raise"], np.number],
            default=np.nan,
        ),
        SarusParameter(
            name="return_train_score",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return model_selection.GridSearchCV(**kwargs)


class sk_best_params(ExternalOpImplementation):
    _transform_id = "sklearn.SK_BEST_PARAMS_"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GridSearchCV,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this = signature.collect_args()[0]
        return this.best_params_


class sk_best_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_BEST_SCORE_"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GridSearchCV,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this = signature.collect_args()[0]
        return this.best_score_



from typing import Any, List, Literal, Optional, Tuple, Union

from numpy.typing import ArrayLike, DTypeLike
import numpy as np

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusParameterArray,
    SarusSignature,
    SarusSignatureValue,
)

from .external_op import ExternalOpImplementation

Casting = Literal["no", "equiv", "safe", "same_kind", "unsafe"]
Order = Literal["K", "A", "C", "F"]


# ------ CONSTRUCTORS ------
class np_array(ExternalOpImplementation):
    _transform_id = "numpy.NP_ARRAY"
    _signature = SarusSignature(
        SarusParameter(
            name="object",
            annotation=ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[DTypeLike],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="order",
            annotation=Order,
            default="K",
        ),
        SarusParameter(
            name="subok",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="ndmin",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="like",
            annotation=Optional[ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["like"] is None:
            del kwargs["like"]
        return np.array(**kwargs)


# ------ FUNCTIONS ------
class np_ceil(ExternalOpImplementation):
    _transform_id = "numpy.NP_CEIL"
    _signature = SarusSignature(
        SarusParameter(
            name="x",
            annotation=ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="out",
            annotation=Optional[ArrayLike],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="where",
            annotation=Optional[Union[bool, ArrayLike]],
            default=True,
        ),
        SarusParameter(
            name="casting",
            annotation=Casting,
            default="same_kind",
        ),
        SarusParameter(
            name="order",
            annotation=Order,
            default="K",
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[DTypeLike],
            default=None,
        ),
        SarusParameter(
            name="subok",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        x, kwargs = signature.collect_kwargs_method()
        return np.ceil(x, **kwargs)


class np_floor(ExternalOpImplementation):
    _transform_id = "numpy.NP_FLOOR"
    _signature = SarusSignature(
        SarusParameter(
            name="x",
            annotation=ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="out",
            annotation=Optional[ArrayLike],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="where",
            annotation=Optional[Union[bool, ArrayLike]],
            default=True,
        ),
        SarusParameter(
            name="casting",
            annotation=Casting,
            default="same_kind",
        ),
        SarusParameter(
            name="order",
            annotation=Order,
            default="K",
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[DTypeLike],
            default=None,
        ),
        SarusParameter(
            name="subok",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        x, kwargs = signature.collect_kwargs_method()
        return np.floor(x, **kwargs)


class np_mean(ExternalOpImplementation):
    _transform_id = "numpy.NP_MEAN"
    _signature = SarusSignature(
        SarusParameter(
            name="a",
            annotation=ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[Union[int, Tuple[int]]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[DTypeLike],
            default=None,
        ),
        SarusParameter(
            name="out",
            annotation=Optional[ArrayLike],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="keepdims",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="where",
            annotation=Optional[Union[bool, ArrayLike]],
            default=np._NoValue,  # type: ignore
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return np.mean(**kwargs)


class np_std(ExternalOpImplementation):
    _transform_id = "numpy.NP_STD"
    _signature = SarusSignature(
        SarusParameter(
            name="a",
            annotation=ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[Union[int, Tuple[int]]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[DTypeLike],
            default=None,
        ),
        SarusParameter(
            name="out",
            annotation=Optional[ArrayLike],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="ddof",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="keepdims",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="where",
            annotation=Optional[Union[bool, ArrayLike]],
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return np.std(**kwargs)


class np_rand(ExternalOpImplementation):
    _transform_id = "numpy.NP_RAND"
    _signature = SarusSignature(
        SarusParameterArray(
            name="size",
            annotation=Optional[Union[int, List[int]]],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        sizes = signature.collect_args()
        return np.random.random_sample(sizes)


class np_ravel_constructor(ExternalOpImplementation):
    _transform_id = "numpy.NP_RAVEL"
    _signature = SarusSignature(
        SarusParameter(
            name="a",
            annotation=ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="order",
            annotation=Order,
            default="C",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return np.ravel(**kwargs)


class np_reshape(ExternalOpImplementation):
    _transform_id = "numpy.NP_RESHAPE"
    _signature = SarusSignature(
        SarusParameter(
            name="a",
            annotation=ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="newshape",
            annotation=Optional[Union[int, Tuple[int]]],
        ),
        SarusParameter(
            name="order",
            annotation=Order,
            default="C",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return np.reshape(**kwargs)


# ------ METHODS ------
class np_astype(ExternalOpImplementation):
    _transform_id = "numpy.NP_ASTYPE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=np.ndarray,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="dtype",
            annotation=DTypeLike,
        ),
        SarusParameter(
            name="order",
            annotation=Order,
            default="K",
        ),
        SarusParameter(
            name="casting",
            annotation=Casting,
            default="unsafe",
        ),
        SarusParameter(
            name="subok",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.astype(**kwargs)


class np_ravel_method(ExternalOpImplementation):
    _transform_id = "numpy.NP_RAVEL_METHOD"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=np.ndarray,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="order",
            annotation=Order,
            default="C",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.ravel(**kwargs)


class np_reshape_method(ExternalOpImplementation):
    _transform_id = "numpy.NP_RESHAPE_METHOD"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=np.ndarray,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="newshape",
            annotation=Optional[Union[int, Tuple[int]]],
        ),
        SarusParameter(
            name="order",
            annotation=Order,
            default="C",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        newshape = kwargs["newshape"]
        order = kwargs["order"]
        return this.reshape(newshape, order=order)



from typing import Any, Callable, Dict, Optional, Union

import numpy as np
import pandas as pd

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusParameterArray,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from shap.maskers import (
        Composite,
        Image,
        Independent,
        Masker,
        Partition,
        Text,
    )
    import shap
except ModuleNotFoundError:
    Independent = Any
    Partition = Any
    Text = Any
    Composite = Any
    Masker = Any
    Text = Any
    Image = Any


# ------ CONSTRUCTORS -------
class shap_masker(ExternalOpImplementation):
    _transform_id = "shap.SHAP_MASKER"
    _signature = SarusSignature()

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.maskers.Masker(**kwargs)


# ------ CONSTRUCTORS -------


class shap_independent(ExternalOpImplementation):
    _transform_id = "shap.SHAP_INDEPENDENT"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=Union[np.ndarray, pd.DataFrame],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="max_samples",
            annotation=Optional[int],
            default=100,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.maskers.Independent(**kwargs)


# ------ Independent METHODS ------


class shap_invariants(ExternalOpImplementation):
    _transform_id = "shap.SHAP_INVARIANTS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Independent,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="x",
            annotation=Union[np.ndarray, pd.DataFrame],
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.invariants(**kwargs)


# ------ CONSTRUCTORS -------
class shap_masker_partition(ExternalOpImplementation):
    _transform_id = "shap.SHAP_MASKER_PARTITION"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=Union[np.ndarray, pd.DataFrame],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="max_samples",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="clustering",
            annotation=str,
            default="correlation",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.maskers.Partition(**kwargs)


# ------ Partition METHODS ------


class shap_partition_invariants(ExternalOpImplementation):
    _transform_id = "shap.SHAP_PARTITION_INVARIANTS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Partition,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="x",
            annotation=np.ndarray,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.invariants(**kwargs)


# ------ CONSTRUCTORS -------
class shap_impute(ExternalOpImplementation):
    _transform_id = "shap.SHAP_IMPUTE"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=Union[np.ndarray, pd.DataFrame, Dict[str, np.ndarray]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="method",
            annotation=str,
            default="linear",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.maskers.Impute(**kwargs)


# ------ CONSTRUCTORS -------


class shap_fixed(ExternalOpImplementation):
    _transform_id = "shap.SHAP_FIXED"
    _signature = SarusSignature()

    def call(self, signature: SarusSignatureValue) -> Any:
        return shap.maskers.Fixed()


# ------ FIXED METHODS ------


class shap_mask_shapes(ExternalOpImplementation):
    _transform_id = "shap.SHAP_FIXED_MASK_SHAPES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Text,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[np.ndarray, pd.DataFrame, Dict[str, np.ndarray]],
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, X = signature.collect_args()
        return this.mask_shapes(X)


# ------ CONSTRUCTORS -------
class shap_composite(ExternalOpImplementation):
    _transform_id = "shap.SHAP_COMPOSITE"
    _signature = SarusSignature(
        SarusParameterArray(
            name="maskers",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        maskers = signature.collect_args()
        return shap.maskers.Composite(*maskers)


# ------ Composite METHODS ------


class shap_data_transform(ExternalOpImplementation):
    _transform_id = "shap.SHAP_DATA_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Composite,
            condition=DATASPEC | STATIC,
        ),
        SarusParameterArray(
            name="args",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, args = signature.collect_args()
        return this.data_transform(*args)


# ------ CONSTRUCTORS -------
class shap_fixed_composite(ExternalOpImplementation):
    _transform_id = "shap.SHAP_FIXED_COMPOSITE"
    _signature = SarusSignature(
        SarusParameter(
            name="masker",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.maskers.FixedComposite(**kwargs)


# ------ CONSTRUCTORS -------


class shap_output_composite(ExternalOpImplementation):
    _transform_id = "shap.SHAP_OUTPUT_COMPOSITE"
    _signature = SarusSignature(
        SarusParameter(
            name="masker",
            annotation=Masker,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="model",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.maskers.OutputComposite(**kwargs)


# ------- CONSTRUCTORS -------
class shap_text_masker(ExternalOpImplementation):
    _transform_id = "shap.SHAP_TEXT"
    _signature = SarusSignature(
        SarusParameter(
            name="tokenizer",
            annotation=Optional[Union[Callable, None]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="mask_token",
            annotation=Optional[Union[str, int, None]],
            default=None,
        ),
        SarusParameter(
            name="collapse_mask_token",
            annotation=Optional[Union[bool, str]],
            default="auto",
        ),
        SarusParameter(
            name="output_type",
            annotation=Optional[str],
            default="string",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.maskers.Text(**kwargs)


# ------ METHODS ------
class shap_clustering(ExternalOpImplementation):
    _transform_id = "shap.SHAP_CLUSTERING"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Text,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="s",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.clustering(**kwargs)


class shap_data_text_transform(ExternalOpImplementation):
    _transform_id = "shap.SHAP_DATA_TEXT_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Text,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="s",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.data_transform(**kwargs)


class shap_feature_names(ExternalOpImplementation):
    _transform_id = "shap.SHAP_FEATURE_NAMES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Text,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="s",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.feature_names(**kwargs)


class shap_text_invariants(ExternalOpImplementation):
    _transform_id = "shap.SHAP_TEXT_INVARIANTS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Text,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="s",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.invariants(**kwargs)


class shap_mask_text_shapes(ExternalOpImplementation):
    _transform_id = "shap.SHAP_MASK_TEXT_SHAPES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Text,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="s",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.mask_shapes(**kwargs)


class shap_shape(ExternalOpImplementation):
    _transform_id = "shap.SHAP_SHAPE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Text,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="s",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.shape(**kwargs)


class shap_token_segments(ExternalOpImplementation):
    _transform_id = "shap.SHAP_TOKEN_SEGMENTS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Text,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="s",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.token_segments(**kwargs)


# ------ CONSTRUCTORS -------
class shap_image(ExternalOpImplementation):
    _transform_id = "shap.SHAP_IMAGE"
    _signature = SarusSignature(
        SarusParameter(
            name="mask_value",
            annotation=Union[np.array, str],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[tuple],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.maskers.Image(**kwargs)


# ------ ImageMasker METHODS ------


class shap_build_partition_tree(ExternalOpImplementation):
    _transform_id = "shap.SHAP_BUILD_PARTITION_TREE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Image,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.build_partition_tree(**kwargs)


class shap_inpaint(ExternalOpImplementation):
    _transform_id = "shap.SHAP_INPAINT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Image,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="x",
            annotation=Union[pd.Series, pd.DataFrame, np.array],
        ),
        SarusParameter(
            name="mask",
            annotation=Union[pd.Series, pd.DataFrame, np.array],
        ),
        SarusParameter(
            name="method",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.inpaint(**kwargs)



from typing import Any, Dict, Iterable, List, Optional

from sarus_data_spec.dataspec_validator.signature import (
    SarusBoundSignature,
    SarusParameter,
    SarusParameterArray,
    SarusParameterMapping,
    SarusSignature,
    SarusSignatureValue,
)
from sarus_data_spec.dataspec_validator.typing import PUPKind

from .external_op import ExternalOpImplementation


class add(ExternalOpImplementation):
    _transform_id = "std.ADD"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this + other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class sub(ExternalOpImplementation):
    _transform_id = "std.SUB"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
        name="substract",
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this - other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class rsub(ExternalOpImplementation):
    _transform_id = "std.RSUB"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return other - this

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class mul(ExternalOpImplementation):
    _transform_id = "std.MUL"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
        name="multiply",
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this * other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class div(ExternalOpImplementation):
    _transform_id = "std.DIV"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this / other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class rdiv(ExternalOpImplementation):
    _transform_id = "std.RDIV"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return other / this

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class invert(ExternalOpImplementation):
    _transform_id = "std.INVERT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return ~this

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class length(ExternalOpImplementation):
    _transform_id = "std.LEN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return len(this)


class getitem(ExternalOpImplementation):
    _transform_id = "std.GETITEM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="key",
            annotation=Any,
        ),
        name=_transform_id,
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, key = signature.collect_args()
        return this[key]

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.PUP


class setitem(ExternalOpImplementation):
    _transform_id = "std.SETITEM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="key",
            annotation=Any,
        ),
        SarusParameter(
            name="value",
            annotation=Any,
        ),
        name=_transform_id,
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, key, value = signature.collect_args()
        this[key] = value
        return this

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        token_value = bound_signature["value"].pup_token()
        token_previous = bound_signature["this"].pup_token()
        if token_value == token_previous:
            return PUPKind.TOKEN_PRESERVING
        return PUPKind.NOT_PUP


class greater_than(ExternalOpImplementation):
    _transform_id = "std.GT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this > other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class greater_equal(ExternalOpImplementation):
    _transform_id = "std.GE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this >= other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class lower_than(ExternalOpImplementation):
    _transform_id = "std.LT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this < other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class lower_equal(ExternalOpImplementation):
    _transform_id = "std.LE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this <= other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class not_equal(ExternalOpImplementation):
    _transform_id = "std.NE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this != other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class equal(ExternalOpImplementation):
    _transform_id = "std.EQ"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this == other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class neg(ExternalOpImplementation):
    _transform_id = "std.NEG"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return -this

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class pos(ExternalOpImplementation):
    _transform_id = "std.POS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return +this

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class _abs(ExternalOpImplementation):
    _transform_id = "std.ABS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return abs(this)

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class _round(ExternalOpImplementation):
    _transform_id = "std.ROUND"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return round(this)

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class modulo(ExternalOpImplementation):
    _transform_id = "std.MOD"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this % other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class rmodulo(ExternalOpImplementation):
    _transform_id = "std.RMOD"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return other % this

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class _or(ExternalOpImplementation):
    _transform_id = "std.OR"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this | other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class ror(ExternalOpImplementation):
    _transform_id = "std.ROR"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return other | this

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class _and(ExternalOpImplementation):
    _transform_id = "std.AND"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return this & other

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class rand(ExternalOpImplementation):
    _transform_id = "std.RAND"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, other = signature.collect_args()
        return other & this

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        return PUPKind.TOKEN_PRESERVING


class _int(ExternalOpImplementation):
    _transform_id = "std.INT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return int(this)


class _float(ExternalOpImplementation):
    _transform_id = "std.FLOAT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return float(this)


class _list(ExternalOpImplementation):
    _transform_id = "std.LIST"
    _signature = SarusSignature(
        SarusParameterArray(
            name="elem",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        elems = signature.collect_args()
        return list(elems)


class _dict(ExternalOpImplementation):
    _transform_id = "std.DICT"
    _signature = SarusSignature(
        SarusParameterMapping(
            name="elem",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        _, elems = signature.collect()
        return dict(**elems)


class _slice(ExternalOpImplementation):
    _transform_id = "std.SLICE"
    _signature = SarusSignature(
        SarusParameter(
            name="start",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="stop",
            annotation=int,
            default=-1,
        ),
        SarusParameter(
            name="step",
            annotation=Optional[int],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (start, stop, step) = signature.collect_args()
        return slice(start, stop, step)


class _set(ExternalOpImplementation):
    _transform_id = "std.SET"
    _signature = SarusSignature(
        SarusParameterArray(
            name="elem",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        elems = signature.collect_args()
        return set(elems)


class _tuple(ExternalOpImplementation):
    _transform_id = "std.TUPLE"
    _signature = SarusSignature(
        SarusParameterArray(
            name="elem",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        elems = signature.collect_args()
        return tuple(elems)


class _string(ExternalOpImplementation):
    _transform_id = "std.STRING"
    _signature = SarusSignature(
        SarusParameterArray(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return str(this)


class _bool(ExternalOpImplementation):
    _transform_id = "std.BOOL"
    _signature = SarusSignature(
        SarusParameterArray(
            name="this",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return bool(this)


class keys(ExternalOpImplementation):
    _transform_id = "std.KEYS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Dict,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return list(this.keys())


class values(ExternalOpImplementation):
    _transform_id = "std.VALUES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Dict,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return list(this.values())


class sudo(ExternalOpImplementation):
    _transform_id = "std.SUDO"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
        ),
        name="sudo",
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this


class extend(ExternalOpImplementation):
    _transform_id = "std.EXTEND"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=List,
        ),
        SarusParameter(
            name="other",
            annotation=List,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, other) = signature.collect_args()
        this.extend(other)
        return this


class append(ExternalOpImplementation):
    _transform_id = "std.APPEND"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=List,
        ),
        SarusParameter(
            name="other",
            annotation=Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, other) = signature.collect_args()
        this.append(other)
        return this


class pop(ExternalOpImplementation):
    _transform_id = "std.POP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=List,
        ),
        SarusParameter(
            name="index",
            annotation=int,
            default=-1,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, index) = signature.collect_args()
        return this.pop(index)


class split(ExternalOpImplementation):
    _transform_id = "std.SPLIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
        SarusParameter(
            name="separator",
            annotation=str,
            default=" ",
        ),
        SarusParameter(
            name="maxsplit",
            annotation=int,
            default=-1,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, separator, maxsplit = signature.collect_args()
        return this.split(separator, maxsplit)


class join(ExternalOpImplementation):
    _transform_id = "std.JOIN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
        SarusParameter(
            name="iterable",
            annotation=Iterable,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, iterable = signature.collect_args()
        return this.join(iterable)


class capitalize(ExternalOpImplementation):
    _transform_id = "std.CAPITALIZE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.capitalize()


class casefold(ExternalOpImplementation):
    _transform_id = "std.CASEFOLD"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.casefold()


class center(ExternalOpImplementation):
    _transform_id = "std.CENTER"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
        SarusParameter(
            name="width",
            annotation=int,
        ),
        SarusParameter(name="fillchar", annotation=str, default=" "),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, width, fillchar) = signature.collect_args()
        return this.center(width, fillchar)


class expandtabs(ExternalOpImplementation):
    _transform_id = "std.EXPANDTABS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
        SarusParameter(
            name="tabsize",
            annotation=int,
            default=8,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, tabsize) = signature.collect_args()
        return this.expandtabs(tabsize)


class lower(ExternalOpImplementation):
    _transform_id = "std.LOWER"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.lower()


class upper(ExternalOpImplementation):
    _transform_id = "std.UPPER"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.upper()


class lstrip(ExternalOpImplementation):
    _transform_id = "std.LSTRIP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
        SarusParameter(
            name="chars",
            annotation=str,
            default=" ",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, chars) = signature.collect_args()
        return this.lstrip(chars)


class rstrip(ExternalOpImplementation):
    _transform_id = "std.RSTRIP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
        SarusParameter(
            name="chars",
            annotation=str,
            default=" ",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, chars) = signature.collect_args()
        return this.rstrip(chars)


class strip(ExternalOpImplementation):
    _transform_id = "std.STRIP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
        SarusParameter(
            name="chars",
            annotation=str,
            default=" ",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, chars) = signature.collect_args()
        return this.strip(chars)


class replace(ExternalOpImplementation):
    _transform_id = "std.REPLACE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
        SarusParameter(
            name="old",
            annotation=str,
        ),
        SarusParameter(
            name="new",
            annotation=str,
        ),
        SarusParameter(
            name="count",
            annotation=int,
            default=-1,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, old, new, count) = signature.collect_args()
        return this.replace(old, new, count)


class splitlines(ExternalOpImplementation):
    _transform_id = "std.SPLITLINES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=str,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.splitlines()



from enum import Enum

from sarus_data_spec import typing as st

DATA = "sarus_data"
PU_COLUMN = "sarus_privacy_unit"
PID_COLUMN = PU_COLUMN
WEIGHTS = "sarus_weights"
PUBLIC = "sarus_is_public"

# constants used in protection stored in each struct
NON_EMPTY_PRIVACY_UNIT_TRACKING_PATHS = "non_zero_protected_values"
STRUCT_KIND = "merge_paths"
TO_MERGE = "fks_for_merging"

# protection constants for paths in types
LIST_VALUES = "sarus_list_values"
ARRAY_VALUES = "sarus_array_values"
OPTIONAL_VALUE = "sarus_optional_value"
CONSTRAINED_VALUE = "sarus_constrained_value"


class StructKind(Enum):
    HAS_PU = "0"
    NO_PU = "1"
    TO_MERGE = "2"


# constants for type properties
TEXT_MIN_LENGTH = "min_length"
TEXT_MAX_LENGTH = "max_length"
TEXT_CHARSET = "text_char_set"
TEXT_EXACT_CHARSET = "FullUserInput"
TEXT_ALPHABET_NAME = "text_alphabet_name"
SQL_MAPPING = "sql_mapping"
FLOAT_DISTRIBUTION = "distribution_model"
FLOAT_DIST_PARAMS = "parameters"
MAX_MAX_MULT = "max_max_multiplicity"
MULTIPLICITY = "multiplicity"
RECOMPUTE_TYPE_RANGES = "recompute_type_ranges"

# constants for dataset properties
DATASET_SLUGNAME = "slugname"

# constants for schema properties
PRIMARY_KEYS = "primary_keys"
FOREIGN_KEYS = "foreign_keys"

# names when transforming datetime type in struct
DATETIME_YEAR = "year"
DATETIME_MONTH = "month"
DATETIME_DAY = "day"
DATETIME_HOUR = "hour"
DATETIME_MINUTES = "minutes"
DATETIME_SECONDS = "seconds"


# sql, to_sql, push_sql status
TO_SQL_TASK = "sql_preparation"
SQL_TASK = "sql"
PUSH_SQL_TASK = "push_sql"

# Big Data Status
BIG_DATA_TASK = "big_data_dataset"
BIG_DATA_THRESHOLD = "threshold"
IS_BIG_DATA = "is_big_data"
DATASET_N_LINES = "dataset_n_lines"
DATASET_N_BYTES = "dataset_n_bytes"
THRESHOLD_TYPE = "threshold_type"
SAMPLE_SIZE_N_LINES = "sample_size_n_lines"

# Caching Status
TO_PARQUET_TASK = "to_parquet"
CACHE = TO_PARQUET_TASK
CACHE_PATH = "path"
COMPUTATION_QUEUED = "computation_queued"
TO_SQL_CACHING_TASK = "to_sql_caching"
SQL_CACHING_URI = "sql_caching_uri"
TABLE_MAPPING = "table_mapping"
EXTENDED_TABLE_MAPPING = "extended_table_mapping"

# Caching infos for scalar
CACHE_TYPE = "cache_type"
CACHE_PROTO = "cache_proto"
SCALAR_TASK = "scalar_value"


class ScalarCaching(Enum):
    PICKLE = "pickle"
    PROTO = "protobuf"
    TRAIN_STATE = "train_state"


# Attributes Status
SCHEMA_TASK = "schema"
SIZE_TASK = "size"
MULTIPLICITY_TASK = "multiplicity"
BOUNDS_TASK = "bounds"
MARGINALS_TASK = "marginals"
ARROW_TASK = "arrow"
PRIVACY_UNIT_TRACKING_TASK = "privacy_unit_tracking_task"
USER_SETTINGS_TASK = "user_settings_task"
PUBLIC_TASK = "public_task"
CACHE_SCALAR_TASK = "cache_scalar"
QB_TASK = "query_builder"
LINKS_TASK = "links_statistics"
SYNTHETIC_TASK = "synthetic"
PROCESSING_INFO = "processing_info"


# Privacy
PUP_TOKEN = "pup_token"
NO_TOKEN = "no_token"
IS_PUBLIC = "is_public"
IS_SYNTHETIC = "is_synthetic"
PRIVACY_LIMIT = "privacy_limit"
CONSTRAINT_KIND = "constraint_kind"
BEST_ALTERNATIVE = "best_alternative"
SALT = "salt"

# QUERYBUILDER
QUERIES = "queries"

# Attributes names
PRIVATE_QUERY = "private_query"
IS_REMOTE = "is_remote"
VARIANT_UUID = "variant_uuid"
RELATIONSHIP_SPEC = "relationship_spec"
VALIDATED_TYPE = "validated_type"
IS_VALID = "is_valid"

# SYNTHETIC DATA
MODEL_PROPERTIES = "model_properties"
SYNTHETIC_MODEL = "sarus_synthetic_model"


class SyntheticDataSettings:
    """Namespace for SD generation settings"""

    BATCH_SIZE = 64
    EPOCHS = 10


# map between sqlalchemy dialect names to sarus notion of dialects
SQLALCHEMY_DIALECT_MAP = {
    "postgresql": st.SQLDialect.POSTGRES,
    "mongodb": st.SQLDialect.MONGODB,
    "mssql": st.SQLDialect.SQL_SERVER,
    "mysql": st.SQLDialect.MY_SQL,
    "sqlite": st.SQLDialect.SQLLITE,
    "oracle": st.SQLDialect.ORACLE,
    "bigquery": st.SQLDialect.BIG_QUERY,
    "redshift": st.SQLDialect.REDSHIFT,
    "hive": st.SQLDialect.HIVE,
    "databricks": st.SQLDialect.DATABRICKS,
}

# Possible values
POSSIBLE_VALUES_LENGTH = "possible_values_length"
POSSIBLE_VALUES = "possible_values"

# Model checkpoints
TRAIN_STATE = "state"
MODEL_NAME = "model_name"
PROMPT = "prompt"

# relationship spec
SHARED_ID_SPACES = "shared-id-spaces"



from math import ceil
from typing import Union
import logging
import typing as t

import numpy as np
import pandas as pd

from sarus_data_spec.constants import MAX_MAX_MULT
from sarus_data_spec.dataspec_validator.signature import SarusBoundSignature
from sarus_data_spec.manager.ops.processor.external.pandas.pandas_dp import (
    get_budget,
)
from sarus_data_spec.protobuf.utilities import unwrap
import sarus_data_spec.typing as st

from ..external_op import DEFAULT_MAX_MAX_MULT, ExternalOpImplementation

GroupBy = Union[
    pd.core.groupby.DataFrameGroupBy,
    pd.core.groupby.SeriesGroupBy,
]

logger = logging.getLogger(__name__)

try:
    from sarus_data_spec.sarus_statistics.ops.bounds.op import BoundOp
    from sarus_data_spec.sarus_statistics.ops.histograms.op import (
        GroupByCountOp,
    )
    from sarus_data_spec.sarus_statistics.ops.max_multiplicity.op import (
        MaxMultiplicityOp,
    )
    from sarus_data_spec.sarus_statistics.ops.mean.op import GroupByMeanOp
    from sarus_data_spec.sarus_statistics.ops.median.op import GroupbyMedianOp
    from sarus_data_spec.sarus_statistics.ops.std.op import GroupByStdOp
    from sarus_data_spec.sarus_statistics.ops.sum.op import GroupBySumOp
    from sarus_data_spec.sarus_statistics.ops.tau_thresholding.op import (
        TauThresholdingOp,
    )
except ModuleNotFoundError as e_groupby_dp:
    if "sarus_data_spec.sarus_statistics" not in str(e_groupby_dp):
        raise

try:
    from sarus_differential_privacy.query import ComposedPrivateQuery
except ModuleNotFoundError as e_groupby_dp:
    if "sarus_differential_privacy" not in str(e_groupby_dp):
        raise

try:
    from sarus_data_spec.sarus_query_builder.builders.bounds_builder import (
        simple_bounds_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.composed_builder import (
        simple_composed_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.max_multiplicity_builder import (  # noqa : E501
        simple_max_multiplicity_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.mean_builder import (
        mean_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.median_builder import (
        median_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.standard_mechanisms_builder import (  # noqa : E501
        laplace_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.std_builder import (
        std_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.sum_builder import (
        sum_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.tau_thresholding_builder import (  # noqa : E501
        tau_threshold_builder_delta,
    )
    from sarus_data_spec.sarus_query_builder.core.core import (
        OptimizableQueryBuilder,
    )
    from sarus_data_spec.sarus_query_builder.protobuf.query_pb2 import (
        GenericTask,
        Query,
    )
except ModuleNotFoundError as e_groupby_dp:
    if "sarus" not in str(e_groupby_dp):
        raise
    OptimizableQueryBuilder = t.Any  # type: ignore

NUMERIC_TYPES = ("integer", "float", "boolean")

DEFAULT_DELTA = 0.01


def has_budget(signature: SarusBoundSignature) -> bool:
    """Retrieve (epsilon, delta) from signature"""
    return "budget" in signature


class pd_count_groupby_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_COUNT_GROUPBY_DP"
    _non_dp_equivalent_id = "pandas.PD_COUNT_GROUPBY"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        return True

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()
        if "budget" in signature:
            _, delta = await get_budget(signature)
        else:
            delta = DEFAULT_DELTA
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )

        n_cols = len(parent_schema.data_type().children())

        # add tau thresholding
        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_counts_builder = laplace_builder(
            parent_ds, Query(laplace_mechanism=Query.LaplaceMechanism())
        )

        # tau thresholding builder
        tau_builder = tau_threshold_builder_delta(parent_ds, delta)

        builders_list: t.List[OptimizableQueryBuilder] = [
            max_mult_builder,
            tau_builder,
        ]
        builders_list.extend(n_cols * [column_counts_builder])

        # add another builder
        builder = simple_composed_builder(parent_ds, builders_list)
        return builder

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframegroupby: pd.core.groupby.DataFrameGroupBy = (
            await parent_ds.async_to(pd.core.groupby.DataFrameGroupBy)
        )
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()

        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget_value[0].epsilon
        random_generator = np.random.default_rng(abs(seed_value))
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, tau_thresholding_task, shape_task = (
            tasks[0],
            tasks[1],
            tasks[2:],
        )

        epsilon = budget_value[0].epsilon
        random_generator = np.random.default_rng(abs(seed_value))

        # Compute DP value

        # we supposed: if max_mult_task has no compute then max mult is 1
        if max_mult_task.HasField("compute"):
            max_mul = MaxMultiplicityOp(
                parent_ds,
                epsilon,  # parameter for quantiles
                max_mult_task.compute.noise_user_count,
                max_mult_task.compute.noise_multiplicity,
                max_mult_task.compute.max_max_multiplicity,
            ).value(random_generator)
        else:
            max_mul = 1

        # keys values from synthetic data
        # syn_parent = parent_ds.variant(st.ConstraintKind.SYNTHETIC)
        # dataframegroupby_synth = await
        # syn_parent.async_to(pd.core.groupby.DataFrameGroupBy)
        # keys_values = [key for key,_ in dataframegroupby_synth]

        # keys values from tau thresholding
        # work with only one index name for the group by
        keys_name = dataframegroupby.keys
        dataset_above_tau_threshold = TauThresholdingOp(
            parent_ds,
            tau_thresholding_task.parameters["epsilon_tau_thresholding"],
            tau_thresholding_task.parameters["delta_tau_thresholding"],
        ).value(keys_name, max_mul, random_generator)
        keys_values = list(dataset_above_tau_threshold.index)

        dataframes_with_counts = {}
        for index, column_name in enumerate(dataframegroupby.nth(0).columns):
            count_parameters = shape_task[index]
            dataframes_with_counts[column_name] = GroupByCountOp(
                parent_ds,
                count_parameters.parameters["noise"],
            ).value(column_name, max_mul, random_generator, keys_values)

        for col_name, df in dataframes_with_counts.items():
            df.columns = [col_name]

        if len(dataframes_with_counts.keys()) == 1:
            return list(dataframes_with_counts.values())[0]
        else:
            return pd.concat(dataframes_with_counts.values(), axis=1)


class pd_mean_groupby_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEAN_GROUPBY_DP"
    _non_dp_equivalent_id = "pandas.PD_MEAN_GROUPBY"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`numeric_only` is `True`"""
        numeric_only = signature["numeric_only"].static_value()
        return numeric_only is True

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()
        if "budget" in signature:
            _, delta = await get_budget(signature)
        else:
            delta = DEFAULT_DELTA

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_bounds_mean_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    mean_builder(parent_ds, Query(mean=Query.Mean())),
                ],
            )
        )
        tau_builder = tau_threshold_builder_delta(parent_ds, delta)
        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder, tau_builder]
            + n_cols * [column_bounds_mean_builder],
        )
        return builder

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframegroupby: pd.core.groupby.DataFrameGroupBy = (
            await parent_ds.async_to(pd.core.groupby.DataFrameGroupBy)
        )
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()

        _, tasks = await self.private_queries_and_task(signature)

        tasks = [unwrap(subtask) for subtask in tasks.subtasks]

        max_mult_task, tau_thresholding_task, mean_tasks = (
            tasks[0],
            tasks[1],
            tasks[2:],
        )

        # Compute DP value
        random_generator = np.random.default_rng(abs(seed_value))
        epsilon = budget_value[0].epsilon

        # we supposed: if max_mult_task has no compute then max mult is 1
        if max_mult_task.HasField("compute"):
            max_mul = MaxMultiplicityOp(
                parent_ds,
                epsilon,  # parameter for quantiles
                max_mult_task.compute.noise_user_count,
                max_mult_task.compute.noise_multiplicity,
                max_mult_task.compute.max_max_multiplicity,
            ).value(random_generator)
        else:
            max_mul = 1

        keys_name = dataframegroupby.keys

        dataset_above_tau_threshold = TauThresholdingOp(
            parent_ds,
            tau_thresholding_task.parameters["epsilon_tau_thresholding"],
            tau_thresholding_task.parameters["delta_tau_thresholding"],
        ).value(keys_name, max_mul, random_generator)
        keys_values = list(dataset_above_tau_threshold.index)

        dataframes_with_dp_means = {}
        for index, column_name in enumerate(dataframegroupby.nth(0).columns):
            subtasks = [unwrap(task) for task in mean_tasks[index].subtasks]

            bounds_parameters, mean_parameters = (
                t.cast(GenericTask, subtasks[0]),
                t.cast(GenericTask, subtasks[1]),
            )
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof("type") == "optional":
                dataframes_with_dp_means[column_name] = dataframegroupby.apply(
                    lambda x: np.nan
                ).to_frame(name="Mean")
                continue
            if column_type.WhichOneof("type") not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters["noise"],
            ).value(column_name, max_mul, random_generator)
            mean_op = GroupByMeanOp(
                parent_ds,
                mean_parameters.parameters["noise"],
            )
            dataframes_with_dp_means[column_name] = mean_op.value(
                column_name, max_mul, bounds, random_generator, keys_values
            )

        for col_name, df in dataframes_with_dp_means.items():
            df.columns = [col_name]

        if len(dataframes_with_dp_means.keys()) == 1:
            return list(dataframes_with_dp_means.values())[0]
        else:
            return pd.concat(dataframes_with_dp_means.values(), axis=1)


class pd_median_groupby_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEDIAN_GROUPBY_DP"
    _non_dp_equivalent_id = "pandas.PD_MEDIAN_GROUPBY"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`numeric_only` is `True`"""
        numeric_only = signature["numeric_only"].static_value()
        return numeric_only is True

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()
        if "budget" in signature:
            _, delta = await get_budget(signature)
        else:
            delta = DEFAULT_DELTA
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        tau_builder = tau_threshold_builder_delta(parent_ds, delta)

        column_bounds_median_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    median_builder(parent_ds, Query(median=Query.Median())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder, tau_builder]
            + n_cols * [column_bounds_median_builder],
        )
        return builder

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframegroupby: pd.core.groupby.DataFrameGroupBy = (
            await parent_ds.async_to(pd.core.groupby.DataFrameGroupBy)
        )
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()

        _, tasks = await self.private_queries_and_task(signature)
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, tau_thresholding_task, median_tasks = (
            tasks[0],
            tasks[1],
            tasks[2:],
        )

        random_generator = np.random.default_rng(abs(seed_value))
        epsilon = budget_value[0].epsilon

        # Compute DP value

        # we supposed: if max_mult_task has no compute then max mult is 1
        if max_mult_task.HasField("compute"):
            max_mul = MaxMultiplicityOp(
                parent_ds,
                epsilon,  # parameter for quantiles
                max_mult_task.compute.noise_user_count,
                max_mult_task.compute.noise_multiplicity,
                max_mult_task.compute.max_max_multiplicity,
            ).value(random_generator)
        else:
            max_mul = 1

        keys_name = dataframegroupby.keys

        dataset_above_tau_threshold = TauThresholdingOp(
            parent_ds,
            tau_thresholding_task.parameters["epsilon_tau_thresholding"],
            tau_thresholding_task.parameters["delta_tau_thresholding"],
        ).value(keys_name, max_mul, random_generator)
        keys_values = list(dataset_above_tau_threshold.index)

        dataframes_with_dp_median = {}
        for index, column_name in enumerate(dataframegroupby.nth(0).columns):
            subtasks = [unwrap(task) for task in median_tasks[index].subtasks]
            bounds_parameters, median_parameters = (
                t.cast(GenericTask, subtasks[0]),
                t.cast(GenericTask, subtasks[1]),
            )
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof("type") == "optional":
                dataframes_with_dp_median[column_name] = (
                    dataframegroupby.apply(
                        lambda x: np.nan
                    ).to_frame(name="Median")
                )
                continue
            if column_type.WhichOneof("type") not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters["noise"],
            ).value(column_name, max_mul, random_generator)
            median_op = GroupbyMedianOp(
                parent_ds,
                median_parameters.parameters["noise"],
            )
            dataframes_with_dp_median[column_name] = median_op.value(
                column_name, max_mul, bounds, random_generator, keys_values
            )

        for col_name, df in dataframes_with_dp_median.items():
            df.columns = [col_name]

        if len(dataframes_with_dp_median.keys()) == 1:
            return list(dataframes_with_dp_median.values())[0]
        else:
            return pd.concat(dataframes_with_dp_median.values(), axis=1)


class pd_std_groupby_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_STD_GROUPBY_DP"
    _non_dp_equivalent_id = "pandas.PD_STD_GROUPBY"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        return True

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()
        if "budget" in signature:
            _, delta = await get_budget(signature)
        else:
            delta = DEFAULT_DELTA
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )

        tau_builder = tau_threshold_builder_delta(parent_ds, delta)

        column_bounds_std_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    std_builder(parent_ds, Query(std=Query.Std())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder, tau_builder]
            + n_cols * [column_bounds_std_builder],
        )
        return builder

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframegroupby: pd.core.groupby.DataFrameGroupBy = (
            await parent_ds.async_to(pd.core.groupby.DataFrameGroupBy)
        )
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()

        _, tasks = await self.private_queries_and_task(signature)
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]

        max_mult_task, tau_thresholding_task, std_tasks = (
            tasks[0],
            tasks[1],
            tasks[2:],
        )

        epsilon = budget_value[0].epsilon
        random_generator = np.random.default_rng(abs(seed_value))

        # Compute DP value

        # we supposed: if max_mult_task has no compute then max mult is 1
        if max_mult_task.HasField("compute"):
            max_mul = MaxMultiplicityOp(
                parent_ds,
                epsilon,  # parameter for quantiles
                max_mult_task.compute.noise_user_count,
                max_mult_task.compute.noise_multiplicity,
                max_mult_task.compute.max_max_multiplicity,
            ).value(random_generator)
        else:
            max_mul = 1

        keys_name = dataframegroupby.keys
        dataset_above_tau_threshold = TauThresholdingOp(
            parent_ds,
            tau_thresholding_task.parameters["epsilon_tau_thresholding"],
            tau_thresholding_task.parameters["delta_tau_thresholding"],
        ).value(keys_name, max_mul, random_generator)
        keys_values = list(dataset_above_tau_threshold.index)

        dataframes_with_dp_std = {}
        for index, column_name in enumerate(dataframegroupby.nth(0).columns):
            subtasks = [unwrap(task) for task in std_tasks[index].subtasks]
            bounds_parameters, std_parameters = (
                t.cast(GenericTask, subtasks[0]),
                t.cast(GenericTask, subtasks[1]),
            )
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof("type") == "optional":
                dataframes_with_dp_std[column_name] = dataframegroupby.apply(
                    lambda x: np.nan
                ).to_frame(name="Std")
                continue
            if column_type.WhichOneof("type") not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters["noise"],
            ).value(column_name, max_mul, random_generator)
            std_op = GroupByStdOp(
                parent_ds,
                std_parameters.parameters["noise_mean"],
                std_parameters.parameters["noise_square"],
                std_parameters.parameters["noise_count"],
            )
            dataframes_with_dp_std[column_name] = std_op.value(
                column_name, max_mul, bounds, random_generator, keys_values
            )

        for col_name, df in dataframes_with_dp_std.items():
            df.columns = [col_name]

        if len(dataframes_with_dp_std.keys()) == 1:
            return list(dataframes_with_dp_std.values())[0]
        else:
            return pd.concat(dataframes_with_dp_std.values(), axis=1)


class pd_sum_groupby_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_SUM_GROUPBY_DP"
    _non_dp_equivalent_id = "pandas.PD_SUM_GROUPBY"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`numeric_only` is `True`"""
        numeric_only = signature["numeric_only"].static_value()
        return numeric_only is True

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()
        if "budget" in signature:
            _, delta = await get_budget(signature)
        else:
            delta = DEFAULT_DELTA

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        tau_builder = tau_threshold_builder_delta(parent_ds, delta)

        column_bounds_sum_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    sum_builder(parent_ds, Query(sum=Query.Sum())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder, tau_builder]
            + n_cols * [column_bounds_sum_builder],
        )
        return builder

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframegroupby: pd.core.groupby.DataFrameGroupBy = (
            await parent_ds.async_to(pd.core.groupby.DataFrameGroupBy)
        )
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()

        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget_value[0].epsilon
        # Compute DP value
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, tau_thresholding_task, sum_tasks = (
            tasks[0],
            tasks[1],
            tasks[2:],
        )

        random_generator = np.random.default_rng(abs(seed_value))

        # we supposed: if max_mult_task has no compute then max mult is 1
        if max_mult_task.HasField("compute"):
            max_mul = MaxMultiplicityOp(
                parent_ds,
                epsilon,  # parameter for quantiles
                max_mult_task.compute.noise_user_count,
                max_mult_task.compute.noise_multiplicity,
                max_mult_task.compute.max_max_multiplicity,
            ).value(random_generator)
        else:
            max_mul = 1

        keys_name = dataframegroupby.keys
        dataset_above_tau_threshold = TauThresholdingOp(
            parent_ds,
            tau_thresholding_task.parameters["epsilon_tau_thresholding"],
            tau_thresholding_task.parameters["delta_tau_thresholding"],
        ).value(keys_name, max_mul, random_generator)
        keys_values = list(dataset_above_tau_threshold.index)

        dataframes_with_dp_sum = {}
        for index, column_name in enumerate(dataframegroupby.nth(0).columns):
            subtasks = [unwrap(task) for task in sum_tasks[index].subtasks]
            bounds_parameters, sum_parameters = (
                t.cast(GenericTask, subtasks[0]),
                t.cast(GenericTask, subtasks[1]),
            )
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof("type") == "optional":
                dataframes_with_dp_sum[column_name] = dataframegroupby.apply(
                    lambda x: np.nan
                ).to_frame(name="Sum")
                continue
            if column_type.WhichOneof("type") not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters["noise"],
            ).value(column_name, max_mul, random_generator)
            sum_op = GroupBySumOp(
                parent_ds,
                sum_parameters.parameters["noise"],
            )
            dataframes_with_dp_sum[column_name] = sum_op.value(
                column_name, max_mul, bounds, random_generator, keys_values
            )

        for col_name, df in dataframes_with_dp_sum.items():
            df.columns = [col_name]

        if len(dataframes_with_dp_sum.keys()) == 1:
            return list(dataframes_with_dp_sum.values())[0]
        else:
            return pd.concat(dataframes_with_dp_sum.values(), axis=1)



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from numpy.typing import ArrayLike
import numpy as np

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from sklearn import ensemble
    from sklearn.base import BaseEstimator
    from sklearn.model_selection import BaseCrossValidator
except ModuleNotFoundError:
    BaseEstimator = Any
    BaseCrossValidator = Any


class sk_adaboost_classifier(ExternalOpImplementation):
    _transform_id = "sklearn.SK_ADABOOST_CLASSIFIER"
    _signature = SarusSignature(
        SarusParameter(
            name="estimator",
            annotation=Optional[BaseEstimator],
            default=None,
        ),
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=50,
        ),
        SarusParameter(
            name="learning_rate",
            annotation=float,
            default=1.0,
        ),
        SarusParameter(
            name="algorithm",
            annotation=Literal["SAMME", "SAMME.R"],
            default="SAMME.R",
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.AdaBoostClassifier(**kwargs)


class sk_adaboost_regressor(ExternalOpImplementation):
    _transform_id = "sklearn.SK_ADABOOST_REGRESSOR"
    _signature = SarusSignature(
        SarusParameter(
            name="estimator",
            annotation=Optional[BaseEstimator],
            default=None,
        ),
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=50,
        ),
        SarusParameter(
            name="learning_rate",
            annotation=float,
            default=1.0,
        ),
        SarusParameter(
            name="loss",
            annotation=Literal["linear", "square", "exponential"],
            default="linear",
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.AdaBoostRegressor(**kwargs)


class sk_bagging_classifier(ExternalOpImplementation):
    _transform_id = "sklearn.SK_BAGGING_CLASSIFIER"
    _signature = SarusSignature(
        SarusParameter(
            name="estimator",
            annotation=Optional[BaseEstimator],
            default=None,
        ),
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="max_samples",
            annotation=Union[int, float],
            default=1.0,
        ),
        SarusParameter(
            name="max_features",
            annotation=Union[int, float],
            default=1.0,
        ),
        SarusParameter(
            name="bootstrap",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="bootstrap_features",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="oob_score",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.BaggingClassifier(**kwargs)


class sk_bagging_regressor(ExternalOpImplementation):
    _transform_id = "sklearn.SK_BAGGING_REGRESSOR"
    _signature = SarusSignature(
        SarusParameter(
            name="estimator",
            annotation=Optional[BaseEstimator],
            default=None,
        ),
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="max_samples",
            annotation=Union[int, float],
            default=1.0,
        ),
        SarusParameter(
            name="max_features",
            annotation=Union[int, float],
            default=1.0,
        ),
        SarusParameter(
            name="bootstrap",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="bootstrap_features",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="oob_score",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.BaggingRegressor(**kwargs)


class sk_extra_trees_classifier(ExternalOpImplementation):
    _transform_id = "sklearn.SK_EXTRA_TREES_CLASSIFIER"
    _signature = SarusSignature(
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="criterion",
            annotation=Literal["gini", "entropy", "log_loss"],
            default="gini",
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_samples_split",
            annotation=Union[int, float],
            default=2,
        ),
        SarusParameter(
            name="min_samples_leaf",
            annotation=Union[int, float],
            default=1,
        ),
        SarusParameter(
            name="min_weight_fraction_leaf",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_features",
            annotation=Union[Literal["sqrt", "log2"], int, float, None],
            default="sqrt",
        ),
        SarusParameter(
            name="max_leaf_nodes",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_impurity_decrease",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="bootstrap",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="oob_score",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="class_weight",
            annotation=Union[
                Literal["balanced", "balanced_subsample"], dict, list
            ],
            default=None,
        ),
        SarusParameter(
            name="ccp_alpha",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_samples",
            annotation=Optional[Union[int, float]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.ExtraTreesClassifier(**kwargs)


class sk_extra_trees_regressor(ExternalOpImplementation):
    _transform_id = "sklearn.SK_EXTRA_TREES_REGRESSOR"
    _signature = SarusSignature(
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="criterion",
            annotation=Literal[
                "squared_error", "absolute_error", "friedman_mse", "poisson"
            ],
            default="squared_error",
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_samples_split",
            annotation=Union[int, float],
            default=2,
        ),
        SarusParameter(
            name="min_samples_leaf",
            annotation=Union[int, float],
            default=1,
        ),
        SarusParameter(
            name="min_weight_fraction_leaf",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_features",
            annotation=Union[Literal["sqrt", "log2", None], int, float],
            default=1.0,
        ),
        SarusParameter(
            name="max_leaf_nodes",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_impurity_decrease",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="bootstrap",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="oob_score",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="ccp_alpha",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_samples",
            annotation=Optional[Union[int, float]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.ExtraTreesRegressor(**kwargs)


class sk_gradient_boosting_classifier(ExternalOpImplementation):
    _transform_id = "sklearn.SK_GRADIENT_BOOSTING_CLASSIFIER"
    _signature = SarusSignature(
        SarusParameter(
            name="loss",
            annotation=Literal["deviance", "exponential", "log_loss"],
            default="log_loss",
        ),
        SarusParameter(
            name="learning_rate",
            annotation=float,
            default=0.1,
        ),
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="subsample",
            annotation=float,
            default=1.0,
        ),
        SarusParameter(
            name="criterion",
            annotation=Literal["friedman_mse", "squared_error"],
            default="friedman_mse",
        ),
        SarusParameter(
            name="min_samples_split",
            annotation=Union[int, float],
            default=2,
        ),
        SarusParameter(
            name="min_samples_leaf",
            annotation=Union[int, float],
            default=1,
        ),
        SarusParameter(
            name="min_weight_fraction_leaf",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=3,
        ),
        SarusParameter(
            name="min_impurity_decrease",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="init",
            annotation=Optional[Union[BaseEstimator, Literal["zero"]]],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="max_features",
            annotation=Union[
                Literal["auto", "sqrt", "log2"], int, float, None
            ],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="max_leaf_nodes",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="validation_fraction",
            annotation=float,
            default=0.1,
        ),
        SarusParameter(
            name="n_iter_no_change",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="tol",
            annotation=float,
            default=1e-4,
        ),
        SarusParameter(
            name="ccp_alpha",
            annotation=float,
            default=0.0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.GradientBoostingClassifier(**kwargs)


class sk_gradient_boosting_regressor(ExternalOpImplementation):
    _transform_id = "sklearn.SK_GRADIENT_BOOSTING_REGRESSOR"
    _signature = SarusSignature(
        SarusParameter(
            name="loss",
            annotation=Literal[
                "squared_error",
                "absolute_error",
                "huber",
                "quantile",
            ],
            default="squared_error",
        ),
        SarusParameter(
            name="learning_rate",
            annotation=float,
            default=0.1,
        ),
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="subsample",
            annotation=float,
            default=1.0,
        ),
        SarusParameter(
            name="criterion",
            annotation=Literal["friedman_mse", "squared_error"],
            default="friedman_mse",
        ),
        SarusParameter(
            name="min_samples_split",
            annotation=Union[int, float],
            default=2,
        ),
        SarusParameter(
            name="min_samples_leaf",
            annotation=Union[int, float],
            default=1,
        ),
        SarusParameter(
            name="min_weight_fraction_leaf",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=3,
        ),
        SarusParameter(
            name="min_impurity_decrease",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="init",
            annotation=Optional[Union[BaseEstimator, Literal["zero"]]],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="max_features",
            annotation=Optional[
                Union[Literal["auto", "sqrt", "log2"], int, float]
            ],
            default=None,
        ),
        SarusParameter(
            name="alpha",
            annotation=float,
            default=0.9,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="max_leaf_nodes",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="validation_fraction",
            annotation=float,
            default=0.1,
        ),
        SarusParameter(
            name="n_iter_no_change",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="tol",
            annotation=float,
            default=1e-4,
        ),
        SarusParameter(
            name="ccp_alpha",
            annotation=float,
            default=0.0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.GradientBoostingRegressor(**kwargs)


class sk_isolation_forest(ExternalOpImplementation):
    _transform_id = "sklearn.SK_ISOLATION_FOREST"
    _signature = SarusSignature(
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="max_samples",
            annotation=Union[Literal["auto"], int, float],
            default="auto",
        ),
        SarusParameter(
            name="contamination",
            annotation=Union[Literal["auto"], float],
            default="auto",
        ),
        SarusParameter(
            name="max_features",
            annotation=Union[int, float],
            default=1.0,
        ),
        SarusParameter(
            name="bootstrap",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.IsolationForest(**kwargs)


class sk_random_forest_classifier(ExternalOpImplementation):
    _transform_id = "sklearn.SK_RANDOM_FOREST_CLASSIFIER"
    _signature = SarusSignature(
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="criterion",
            annotation=Literal["gini", "entropy", "log_loss"],
            default="gini",
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_samples_split",
            annotation=Union[int, float],
            default=2,
        ),
        SarusParameter(
            name="min_samples_leaf",
            annotation=Union[int, float],
            default=1,
        ),
        SarusParameter(
            name="min_weight_fraction_leaf",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_features",
            annotation=Union[Literal["sqrt", "log2", None], int, float],
            default="sqrt",
        ),
        SarusParameter(
            name="max_leaf_nodes",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_impurity_decrease",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="bootstrap",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="oob_score",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="class_weight",
            annotation=Optional[
                Union[
                    Literal["balanced", "balanced_subsample"], Dict, List[Dict]
                ]
            ],
            default=None,
        ),
        SarusParameter(
            name="ccp_alpha",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_samples",
            annotation=Optional[Union[int, float]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.RandomForestClassifier(**kwargs)


class sk_random_forest_regressor(ExternalOpImplementation):
    _transform_id = "sklearn.SK_RANDOM_FOREST_REGRESSOR"
    _signature = SarusSignature(
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="criterion",
            annotation=Literal[
                "squared_error", "absolute_error", "friedman_mse", "poisson"
            ],
            default="squared_error",
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_samples_split",
            annotation=Union[int, float],
            default=2,
        ),
        SarusParameter(
            name="min_samples_leaf",
            annotation=Union[int, float],
            default=1,
        ),
        SarusParameter(
            name="min_weight_fraction_leaf",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_features",
            annotation=Optional[Union[Literal["sqrt", "log2"], int, float]],
            default=1.0,
        ),
        SarusParameter(
            name="max_leaf_nodes",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_impurity_decrease",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="bootstrap",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="oob_score",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="warm_start",
            annotation=Optional[bool],
            default=False,
        ),
        SarusParameter(
            name="ccp_alpha",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_samples",
            annotation=Optional[Union[int, float]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.RandomForestRegressor(**kwargs)


class sk_random_trees_embedding(ExternalOpImplementation):
    _transform_id = "sklearn.SK_RANDOM_TREES_EMBEDDING"
    _signature = SarusSignature(
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="max_depth",
            annotation=int,
            default=5,
        ),
        SarusParameter(
            name="min_samples_split",
            annotation=Union[int, float],
            default=2,
        ),
        SarusParameter(
            name="min_samples_leaf",
            annotation=Union[int, float],
            default=1,
        ),
        SarusParameter(
            name="min_weight_fraction_leaf",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_leaf_nodes",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_impurity_decrease",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="sparse_output",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.RandomTreesEmbedding(**kwargs)


class sk_stacking_classifier(ExternalOpImplementation):
    _transform_id = "sklearn.SK_STACKING_CLASSIFIER"
    _signature = SarusSignature(
        SarusParameter(
            name="estimators",
            annotation=List[Tuple[str, BaseEstimator]],
        ),
        SarusParameter(
            name="final_estimator",
            annotation=Optional[BaseEstimator],
            default=None,
        ),
        SarusParameter(
            name="cv",
            annotation=Union[
                int, BaseCrossValidator, Iterable, Literal["prefit"]
            ],
            default=None,
        ),
        SarusParameter(
            name="stack_method",
            annotation=Literal[
                "auto", "predict_proba", "decision_function", "predict"
            ],
            default="auto",
        ),
        SarusParameter(
            name="n_jobs",
            annotation=int,
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="passthrough",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.StackingClassifier(**kwargs)


class sk_stacking_regressor(ExternalOpImplementation):
    _transform_id = "sklearn.SK_STACKING_REGRESSOR"
    _signature = SarusSignature(
        SarusParameter(
            name="estimators",
            annotation=List[Tuple[str, BaseEstimator]],
        ),
        SarusParameter(
            name="final_estimator",
            annotation=Optional[BaseEstimator],
            default=None,
        ),
        SarusParameter(
            name="cv",
            annotation=Optional[Union[int, BaseCrossValidator, Iterable[str]]],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="passthrough",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.StackingRegressor(**kwargs)


class sk_voting_classifier(ExternalOpImplementation):
    _transform_id = "sklearn.SK_VOTING_CLASSIFIER"
    _signature = SarusSignature(
        SarusParameter(
            name="estimators",
            annotation=List[Tuple[str, BaseEstimator]],
            default=None,
        ),
        SarusParameter(
            name="voting",
            annotation=Literal["hard", "soft"],
            default="hard",
        ),
        SarusParameter(
            name="weights",
            annotation=Optional[List[float]],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="flatten_transform",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.VotingClassifier(**kwargs)


class sk_voting_regressor(ExternalOpImplementation):
    _transform_id = "sklearn.SK_VOTING_REGRESSOR"
    _signature = SarusSignature(
        SarusParameter(
            name="estimators",
            annotation=List[Tuple[str, BaseEstimator]],
        ),
        SarusParameter(
            name="weights",
            annotation=Optional[List[float]],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.VotingRegressor(**kwargs)


class sk_hist_gradient_boosting_classifier(ExternalOpImplementation):
    _transform_id = "sklearn.SK_HIST_GRADIENT_BOOSTING_CLASSIFIER"
    _signature = SarusSignature(
        SarusParameter(
            name="loss",
            annotation=Literal[
                "log_loss",
                "auto",
                "binary_crossentropy",
                "categorical_crossentropy",
            ],
            default="log_loss",
        ),
        SarusParameter(
            name="learning_rate",
            annotation=float,
            default=0.1,
        ),
        SarusParameter(
            name="max_iter",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="max_leaf_nodes",
            annotation=Optional[int],
            default=31,
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_samples_leaf",
            annotation=int,
            default=20,
        ),
        SarusParameter(
            name="l2_regularization",
            annotation=float,
            default=0,
        ),
        SarusParameter(
            name="max_bins",
            annotation=int,
            default=255,
        ),
        SarusParameter(
            name="categorical_features",
            annotation=Optional[ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="monotonic_cst",
            annotation=Optional[Union[ArrayLike, Dict]],
            default=None,
        ),
        SarusParameter(
            name="interaction_cst",
            annotation=Optional[
                Union[
                    Literal["pairwise", "no_interaction"],
                    Sequence[Union[List[int], Tuple[int], Set[int]]],
                ]
            ],
            default=None,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="early_stopping",
            annotation=Union[Literal["auto"], bool],
            default="auto",
        ),
        SarusParameter(
            name="scoring",
            annotation=Optional[Union[str, Callable, None]],
            default="loss",
        ),
        SarusParameter(
            name="validation_fraction",
            annotation=Optional[Union[int, float]],
            default=0.1,
        ),
        SarusParameter(
            name="n_iter_no_change",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="tol",
            annotation=float,
            default=1e-7,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="class_weight",
            annotation=Optional[Union[str, dict]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.HistGradientBoostingClassifier(**kwargs)


class sk_hist_gradient_boosting_regressor(ExternalOpImplementation):
    _transform_id = "sklearn.SK_HIST_GRADIENT_BOOSTING_REGRESSOR"
    _signature = SarusSignature(
        SarusParameter(
            name="loss",
            annotation=Literal[
                "squared_error", "absolute_error", "poisson", "quantile"
            ],
            default="squared_error",
        ),
        SarusParameter(
            name="quantile",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="learning_rate",
            annotation=float,
            default=0.1,
        ),
        SarusParameter(
            name="max_iter",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="max_leaf_nodes",
            annotation=Optional[int],
            default=31,
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_samples_leaf",
            annotation=int,
            default=20,
        ),
        SarusParameter(
            name="l2_regularization",
            annotation=float,
            default=0,
        ),
        SarusParameter(
            name="max_bins",
            annotation=int,
            default=255,
        ),
        SarusParameter(
            name="categorical_features",
            annotation=Optional[
                Union[Sequence[Union[bool, int, str]], Sequence[int]]
            ],
            default=None,
        ),
        SarusParameter(
            name="monotonic_cst",
            annotation=Optional[Union[Sequence[int], Dict[int, int]]],
            default=None,
        ),
        SarusParameter(
            name="interaction_cst",
            annotation=Optional[
                Union[
                    Literal["pairwise", "no_interaction"],
                    Sequence[Union[List[int], Tuple[int], Set[int]]],
                ]
            ],
            default=None,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="early_stopping",
            annotation=Union[Literal["auto"], bool],
            default="auto",
        ),
        SarusParameter(
            name="scoring",
            annotation=Optional[Union[str, Callable]],
            default="loss",
        ),
        SarusParameter(
            name="n_iter_no_change",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="tol",
            annotation=float,
            default=1e-7,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return ensemble.HistGradientBoostingRegressor(**kwargs)
